#
# CLI for the agent-sensor
#
# Docs:
# - [Typer](https://typer.tiangolo.com/) - CLI API
# - [Rich](https://rich.readthedocs.io/en/stable/) - Display rich text to terminal
#

import logging
import typer
import requests
import socket
import time
import uvicorn

from enum import Enum
from fastapi import FastAPI
from fastapi.responses import Response
from fastapi_utils.tasks import repeat_every
from rich import print
from typing import Optional
from typing_extensions import Annotated
from pathlib import Path

from ays_agent import get_agent_payload, get_config_path, get_name, get_version, CLIOptions, load_options, save_options, strip_v, AgentException
from ays_agent.stat.cpu import CPUMonitor
from ays_agent.stat.disk import DiskMonitor
from ays_agent.stat.memory import MemoryMonitor
from ays_agent.stat.network import NetworkMonitor

class NodeType(str, Enum):
    machine = "machine"
    service = "service"
    vendor = "vendor"

class HeartbeatLevel(str, Enum):
    warning = "warning"
    error = "error"
    critical = "critical"

class StatusState(str, Enum):
    healthy = "healthy"
    warning = "warning"
    error = "error"
    critical = "critical"

class MonitorResource(str, Enum):
    all = "all"
    cpu = "cpu"
    hdd = "hdd"
    ram = "ram"
    net = "net"

RESOURCE_MONITORS = {
    MonitorResource.cpu: CPUMonitor,
    MonitorResource.hdd: DiskMonitor,
    MonitorResource.ram: MemoryMonitor,
    MonitorResource.net: NetworkMonitor
}

app = typer.Typer(no_args_is_help=True, invoke_without_command=True)

def get_default_server():
    return "https://api.bithead.io:9443/agent/"

def get_hostname():
    return socket.gethostname()

def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{get_name()} v{get_version()}")
        raise typer.Exit()

def send_request(server, json):
    resp = requests.request("POST", server, json=json)
    if resp.status_code != 204:
        print("Failed to make request to @ys server")
        print(resp)
        raise typer.Exit(1)
    else:
        raise typer.Exit()

# FastAPI Service

fastapp = FastAPI()

@fastapp.get("/test/")
async def test():
    return Response(status_code=204)

# Typer app

@app.callback()
def main(
    version: Annotated[bool, typer.Option(
        help="Show the application's version and exit.",
        callback=_version_callback,
        show_default=False
    )] = False,
    server: Annotated[str, typer.Option(
        help=f"Path to the @ys server agent endpoint. Default: {get_default_server()}"
    )] = None,
    org_secret: Annotated[str, typer.Option(
        help="Organization secret. Required to interact with the respective org system graph."
    )] = "",
    interval: Annotated[int, typer.Option(
        help="Interval, in seconds, that the agent will report a value or status. This will make the agent act as a long-running service. If a monitor is used, and the interval is not provided, the default interval will be 5 minutes."
    )] = None,
    port: Annotated[int, typer.Option(
        help="The port the agent listens to when becoming a long running service."
    )] = 9555,
    parent: Annotated[str, typer.Option(
        help="The parent node path this agent will relate to.",
        show_default=False
    )] = None,
    monitor_name: Annotated[str, typer.Option(
        help="The name of the monitor that will be associated to the respective node. The default value will be the host name of the machine.",
        show_default=False
    )] = None,
    child: Annotated[str, typer.Option(
        help="A relative path to a child node that lives under the parent's node path.",
        show_default=False
    )] = None,
    create_child: Annotated[bool, typer.Option(
        help="Create a child node that lives under the parent's node path. The name of this child node will use the same name as the monitor."
    )] = False,
    node_type: Annotated[NodeType, typer.Option(
        help="The type of node to create when a child is created. Ignored when no child node is created.",
        show_default=False
    )] = None,
    managed: Annotated[bool, typer.Option(
        help="Agent is responsible for managing its own configuration.",
        show_default=False
    )] = None,

    heartbeat_timeout: Annotated[int, typer.Option(
        help="The timeout, in seconds, in which the agent will be considered unhealthy if it doesn't report within specified time.",
        show_default=False
    )] = None,
    heartbeat_level: Annotated[HeartbeatLevel, typer.Option(
        help="The alerting level to transition into if hearbeat timeout exceeded.",
        show_default=False
    )] = HeartbeatLevel.critical,

    value: Annotated[float, typer.Option(
        help="Report a single value to report to @ys.",
        show_default=False
    )] = None,
    value_name: Annotated[str, typer.Option(
        help="The respective name for the value.",
        show_default=False
    )] = None,
    value_threshold: Annotated[str, typer.Option(
        help="Thresholds are used to determine if the value is nominal or unhealthy. Options are `<N`, `>N`, `eN`, `neN`, and `N-N` where `N` is the threshold value. e.g. `<30` will trigger a threshold breach if `value` is greater than `30`.",
        show_default=False
    )] = None,

    values: Annotated[float, typer.Option(
        help="Report multiple values to @ys.",
        show_default=False
    )] = None,
    value_names: Annotated[str, typer.Option(
        help="The respective names for each value.",
        show_default=False
    )] = None,
    value_thresholds: Annotated[str, typer.Option(
        help="The threshold used to determine if the respective value is nominal or unhealthy.",
        show_default=False
    )] = None,

    status_message: Annotated[str, typer.Option(
        help="The message as to why the status is changing. Default is an empty string if `status-state` is provided.",
        show_default=False
    )] = None,
    status_state: Annotated[StatusState, typer.Option(
        help="The alerting level to transition into if hearbeat timeout exceeded. Default is `critical`, if only `status-message` is provided.",
        show_default=False
    )] = None,

    # Services
    monitor_resources: Annotated[str, typer.Option(
        help="Monitor system resources. Available options: {list(MonitorResource.__members__}",
        show_default=False
    )] = None,

    monitor_file: Annotated[Optional[Path], typer.Option(
        help="Monitor the contents of a CSV file.",
        show_default=False
    )] = None,

    monitor_program: Annotated[Optional[Path], typer.Option(
        help="Execute a CLI program to derive value(s) to report on.",
        show_default=False
    )] = None,

    write_config: Annotated[bool, typer.Option(
        help="Write all options to configuration file."
    )] = False,
    dry_run: Annotated[bool, typer.Option(
        help="Emit the action that will take place, with the specified parameters, w/o sending data to @ys."
    )] = False,
) -> None:
    # Load options from disk, if any
    options = load_options(get_config_path())
    # Merge options provided
    options.merge(
        org_secret=org_secret,
        server=server,
        interval=interval,
        parent=parent,
        monitor_name=monitor_name,
        child=child,
        create_child=create_child,
        node_type=node_type,
        managed=managed,
        heartbeat_timeout=heartbeat_timeout,
        heartbeat_level=heartbeat_level,
        value=value,
        value_name=value_name,
        value_threshold=value_threshold,
        values=values,
        value_names=value_names,
        value_thresholds=value_thresholds,
        status_message=status_message,
        status_state=status_state,
        monitor_resources=monitor_resources,
        monitor_file=monitor_file,
        monitor_program=monitor_program
    )

    if not monitor_name:
        options.monitor_name = get_hostname()

    # Ensure options are valid. This must happen regardless if CLI options are
    # provided or not as the user may write invalid config to the config file.
    server, msg = get_agent_payload(options)

    # NOTE: Options must be checked before they are written to config.
    if write_config:
        save_options(options)
        print("[green]Saved configuration to disk successfully.[/green]")
        raise typer.Exit()

    if (monitor_resources or monitor_program or monitor_file) and not options.interval:
        # Default is 5 minutes
        options.interval = 60 * 5
    if options.interval is not None and options.interval < 15:
        raise AgentException("Interval provided ({options.interval}) must be 15 seconds or greater")

    if dry_run:
        if monitor_resources:
            print(f"Monitor resources: {monitor_resources} every {options.interval}s")
        elif monitor_program:
            print(f"Monitor program: {monitoring_program} every {options.interval}s")
        elif monitor_file:
            print(f"Monitor file: {monitor_file} every {options.interval}s")
        print(f"Server: [green]{server}[/green]")
        print(msg)
        raise typer.Exit()

    if monitor_resources:
        resources = strip_v(monitor_resources)
        for res in resources:
            if res not in MonitorResource.__members__:
                raise AgentException(f"Invalid monitor resource ({res}). Available options are ({', '.join(MonitorResource.__members__)}).")
        monitor_options = monitor_resources.split(",")
        # Get class defintions of monitors to use
        if MonitorResource.all in monitor_options:
            monitors = RESOURCE_MONITORS.values()
        else:
            monitors = list(map(lambda x: RESOURCE_MONITORS[x], monitor_options))
        # Instantiate and start monitors
        monitors = list(map(lambda x: x(), monitors))
        for m in monitors: m.start()

        @fastapp.on_event("startup")
        @repeat_every(seconds=options.interval)
        async def run_forever() -> None:
            values = []
            for monitor in monitors:
                # NOTE: This doesn't provide thresholds for values. If thresholds
                # are required, use the `com.bithead.template.agent_resources`
                # template.
                values.extend(monitor.get_values(options.interval))
            msg["values"] = values
            send_request(server, msg)

        uvicorn.run(fastapp, host="0.0.0.0", port=port)
    elif monitor_program:
        # TODO: Execute program
        pass
    elif monitor_file:
        # TODO: Monitor file
        pass
    elif options.interval:
        @fastapp.on_event("startup")
        @repeat_every(seconds=options.interval)
        async def run_forever() -> None:
            send_request(server, msg)

        uvicorn.run(fastapp, host="0.0.0.0", port=port)
    else:
        send_request(server, msg)



import psutil

from typing import List

from ays_agent.stat import format_bytes, get_base, get_megabytes

class NetworkMonitor(object):
    def __init__(self):
        self.bytes_sent = 0
        self.bytes_recv = 0

    def start(self):
        """ Start monitoring network traffic. """
        io = psutil.net_io_counters()
        self.bytes_sent, self.bytes_recv = io.bytes_sent, io.bytes_recv

    def get_stats(self, delay: int) -> tuple[int, int, int, int]:
        """ Get network stats since last delay.

        @returns the number of bytes sent, bytes received, upload speed, and
        download speed.
        """
        io = psutil.net_io_counters()
        up_speed, dl_speed = (io.bytes_sent - self.bytes_sent) / delay, (io.bytes_recv - self.bytes_recv) / delay
        self.bytes_sent, self.bytes_recv = io.bytes_sent, io.bytes_recv
        return self.bytes_sent, self.bytes_recv, up_speed, dl_speed

    def get_formatted_stats(self, delay: int) -> tuple[str, str, str, str]:
        """ Get formatted network stats since last delay. """
        base = get_base()
        stats = self.get_stats(delay)
        stats = tuple(map(lambda x: format_bytes(x, base), stats))
        return stats

    def get_values(self, delay: int) -> List[dict]:
        """ Get list of values that represent an `AgentValue`. """
        sent, recv, up, dl = self.get_stats(delay)
        return [
            {"name": "Net Sent/sec", "value": get_megabytes(up)},
            {"name": "Net Received/sec", "value": get_megabytes(dl)}
        ]



from .context import ays_agent

import logging
import os
import pytest

from typer.testing import CliRunner
from unittest.mock import patch

from ays_agent import set_config_path, get_config_path, get_name, get_version, cli, AgentException, load_options, CLIOptions

runner = CliRunner()

def test_version():
    result = runner.invoke(cli.app, ["--version"])
    assert result.exit_code == 0
    assert f"{get_name()} v{get_version()}\n" in result.stdout

def patch_string():
    return "testing"

def run_command(options, call):
    result = runner.invoke(cli.app, options)
    args, _ = call.call_args
    return result.exit_code, args

def call_cli(options, call):
    cli.main(**options)
    args, _ = call.call_args
    return args

def setup_config():
    config_path = os.path.join(os.path.dirname(__file__), "ays-agent")
    if os.path.isfile(config_path):
        os.unlink(config_path)
    set_config_path(config_path)

@pytest.fixture(autouse=True)
def setup_tests():
    # Clean config before each test
    setup_config()

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_send_value(p_send_request):
    req_options = [
        "--org-secret=aaa",
        "--parent=com.unittest.send"
    ]
    # describe: send value; send all value options
    code, args = run_command(req_options + ["--value=5", "--value-name=disk", "--value-threshold=\">10:warning\""], p_send_request)
    assert code == 0
    assert args[0] == cli.get_default_server()
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 5.0, "name": "disk", "threshold": {"above": 10.0, "level": "warning"}}
    }, "it: should send provided values"

    # describe: send only value
    code, args = run_command(req_options + ["--value=5"], p_send_request)
    assert code == 0
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 5.0, "name": "value"}
    }, "it: should send default values"

    # describe: threshold is below
    code, args = run_command(req_options + ["--value=3.2", "--value-threshold=<7.1:error"], p_send_request)
    assert code == 0
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 3.2, "name": "value", "threshold": {"below": 7.1, "level": "error"}}
    }, "it: should send default values"

    # describe: threshold is equal
    code, args = run_command(req_options + ["--value=3.2", "--value-threshold=e2:critical"], p_send_request)
    assert code == 0
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 3.2, "name": "value", "threshold": {"equal": 2, "level": "critical"}}
    }, "it: should send default values"

    # describe: threshold is not equal
    code, args = run_command(req_options + ["--value=3.2", "--value-threshold=ne2:critical"], p_send_request)
    assert code == 0
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 3.2, "name": "value", "threshold": {"nequal": 2, "level": "critical"}}
    }, "it: should send correct values"

    # describe: threshold is range
    code, args = run_command(req_options + ["--value=3.2", "--value-threshold=40.2-50.5"], p_send_request)
    assert code == 0
    # it: should set default threshold value to `critical`
    assert args[1] == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.send"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "value": {"value": 3.2, "name": "value", "threshold": {"outside": {"min": 40.2, "max": 50.5}, "level": "critical"}}
    }, "it: should send correct values"

    args = {
        "org_secret": "aaa",
        "parent": "com.unittest.send",
        "value": "4.1",
        "value_threshold": "b34"
    }

    # describe: provide inavlid threshold type
    with pytest.raises(AgentException, match=r"^Invalid threshold type for value \(b34\)$"):
        cli.main(**args)

    # describe: provide invalid threshold level
    args["value_threshold"] = ">34:super"
    with pytest.raises(AgentException, match=r"^Invalid threshold level \(super\). Available options are \(warning, error, critical\)$"):
        cli.main(**args)

    # describe: invalid value
    args["value"] = "b34"
    args["value_threshold"] = ">34"
    with pytest.raises(ValueError):
        cli.main(**args)

    # describe: invalid min value
    args["value"] = "34"
    args["value_threshold"] = "o34-40"
    with pytest.raises(ValueError):
        cli.main(**args)

    # describe: invalid max value
    args["value"] = "34"
    args["value_threshold"] = "34-o40"
    with pytest.raises(ValueError):
        cli.main(**args)

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_send_values(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.values",
        "values": "4.1,5.0,7.7",
        "value_names": "cpu, ram,hdd",
        "value_thresholds": "<1.3,>8.7:error,"
    }

    # describe: all value options provided; one threshold omitted; name has space
    req, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.values"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "values": [
            {"value": 4.1, "name": "cpu", "threshold": {"below": 1.3, "level": "critical"}},
            {"value": 5.0, "name": "ram", "threshold": {"above": 8.7, "level": "error"}},
            {"value": 7.7, "name": "hdd"}
        ]
    }, "it: should send correct values"

    # describe: invalid number of names
    options["value_names"] = "cpu,ram"
    with pytest.raises(AgentException, match=r"^The number of names \(2\) must match the number of values \(3\) provided$"):
        cli.main(**options)

    # describe: invalid number of thresholds
    options["value_names"] = "cpu,ram,hdd"
    options["value_thresholds"] = "<1.3,>8.7"
    with pytest.raises(AgentException, match=r"^The number of thresholds \(2\) must match the number of values \(3\) provided$"):
        cli.main(**options)

    # describe: value options provided; name missing
    options.pop("value_thresholds")
    options["value_names"] = "cpu,,hdd"
    req, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.values"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "values": [
            {"value": 4.1, "name": "cpu"},
            # it: should set default value to "value" + index
            {"value": 5.0, "name": "value1"},
            {"value": 7.7, "name": "hdd"}
        ]
    }, "it: should send correct values"

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_send_status(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.values",
        "status_message": "This is only a test",
    }

    # describe: provide status; no state provided
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.values"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "status": {
            "message": "This is only a test",
            # it: should return correct default value
            "state": "critical"
        }
    }, "it: should send correct values"

    # describe: provide status state; no status message
    options.pop("status_message")
    options["status_state"] = "warning"
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.values"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "status": {
            # it: should return correct default value
            "message": "",
            "state": "warning"
        }
    }, "it: should send correct values"

    # describe: provide invalid state
    options["status_state"] = "super"
    with pytest.raises(AgentException, match=r"^Invalid status state \(super\). Available options are \(healthy, warning, error, critical\)$"):
        cli.main(**options)

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_create_child(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.create_child",
        "child": "a-child"
    }

    # describe: provide status; no state provided
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.create_child"},
        "relationship": {"type": "child", "monitor_name": "testing", "path": "a-child"},
    }, "it: should send correct values"

    # describe: child name does not match naming convention
    options["child"] = "a child**node"
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.create_child"},
        "relationship": {"type": "child", "monitor_name": "testing", "path": "a-child-node"},
    }, "it: should send correct values"

    # describe: child name has invalid first character
    options["child"] = "0name"
    with pytest.raises(AgentException, match=r"^A node name must start with one of the following characters \(abcdefghijklmnopqrstuvwxyz\)$"):
        cli.main(**options)

    # describe: child name is greater than max length
    options["child"] = "e123456789012345678901234567890"
    assert len(options["child"]) == 31 # sanity
    with pytest.raises(AgentException, match=r"^A node name may not exceed 30 characters$"):
        cli.main(**options)

    # describe: creat child by adopting the monitor name
    options.pop("child")
    options["create_child"] = True
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.create_child"},
        "relationship": {"type": "child", "monitor_name": "testing", "path": "testing"},
    }, "it: should adopt the monitor name for the child node name"

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_agent_type(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.agent_type",
        "create_child": True,
        "node_type": "machine"
    }

    # describe: provide a valid node type
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.agent_type"},
        "relationship": {"type": "child", "monitor_name": "testing", "path": "testing"},
        "type": "machine"
    }, "it: should adopt the monitor name for the child node name"

    # describe: provide invalid node type
    options["node_type"] = "incorrect"
    with pytest.raises(AgentException, match=r"^Invalid agent type \(incorrect\). Available options are \(machine, service, vendor\)$"):
        cli.main(**options)

    # describe: node type provided; not creating a child
    options["node_type"] = "machine"
    options.pop("create_child")
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.agent_type"},
        "relationship": {"type": "parent", "monitor_name": "testing"}
    }, "it: should ignore the node type"

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_managed_node(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.agent_type",
        "create_child": True,
        "managed": True
    }

    # describe: provide managed
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.agent_type"},
        "relationship": {"type": "child", "monitor_name": "testing", "path": "testing"},
        "managed": True
    }, "it: should be a managed node"

@patch("ays_agent.cli.send_request")
@patch("ays_agent.cli.get_hostname", patch_string)
def test_heartbeat(p_send_request):
    options = {
        "org_secret": "aaa",
        "parent": "com.unittest.heartbeat",
        "heartbeat_timeout": 30
    }

    # describe: provide heartbeat config; no level provided
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.heartbeat"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "heartbeat": {
            "timeout": 30,
            # it: should set to default value
            "level": "critical"
        }
    }, "it: should return heartbeat config"

    # describe: provide heartbeat config; level provided
    options["heartbeat_level"] = "warning"
    _, args = call_cli(options, p_send_request)
    assert args == {
        "org_secret": "aaa",
        "parent": {"property": "path", "value": "com.unittest.heartbeat"},
        "relationship": {"type": "parent", "monitor_name": "testing"},
        "heartbeat": {
            "timeout": 30,
            "level": "warning"
        }
    }, "it: should set heartbeat level"

    # describe: provide invalid heartbeat timeout
    options["heartbeat_timeout"] = "o30"
    with pytest.raises(ValueError):
        cli.main(**options)

    # describe: provide invalid heartbeat level
    options["heartbeat_timeout"] = 30
    options["heartbeat_level"] = "incorrect"
    with pytest.raises(AgentException, match=r"^Invalid heartbeat level \(incorrect\). Available options are \(warning, error, critical\)$"):
        cli.main(**options)

@patch("ays_agent.cli.get_hostname", patch_string)
def test_write_config():
    options = [
        "--org-secret=aaa",
        "--parent=com.unittest.test",
        "--create-child",
        "--heartbeat-timeout=30",
        "--heartbeat-level=warning",
        "--write-config"
    ]

    # describe: write config to disk
    result = runner.invoke(cli.app, options)
    assert result.exit_code == 0
    opts = load_options(get_config_path())
    expected = CLIOptions(
        org_secret="aaa",
        server=cli.get_default_server(),
        parent="com.unittest.test",
        monitor_name="testing",
        create_child=True,
        heartbeat_timeout=30,
        heartbeat_level="warning"
    )
    assert opts.__dict__ == expected.__dict__, "it: should save and load config"

    os.unlink(config_path)

def test_write_config():
    # describe: no optionsn are loaded from disk
    opts = load_options(get_config_path())
    expected = CLIOptions(
        org_secret="",
        server="",
        parent="",
        monitor_name="",
    )
    assert opts.__dict__ == expected.__dict__, "it: should return empty options"



import sys

from typing import Union

# We (as in society) can't decide if the base is 1000 or 1024. HD manufacturer's
# prefer 1000, while IT prefers 1024. This uses the respective base depending
# on the platform's preference. Windows & Linux = 1024. macOS = 1000.
MACOS_BASE = 1000
OTHER_BASE = 1024

def get_platform():
    return sys.platform.lower()

def is_darwin():
    return get_platform() == "darwin"

def is_windows():
    return get_platform() == "windows"

def get_base() -> int:
    """ Returns respective used for computation of byte-sizes given the
    respective platform the agent is running on."""
    if is_darwin():
        return MACOS_BASE
    else:
        return OTHER_BASE

def get_default_mountpoint():
    if is_darwin():
        # TODO: This may be a different name on <=Catalina. I can't get a
        # definitive answer.
        return "/System/Volumes/Data"
    else:
        # TODO: Determine default root path for Linux & Windows
        # This may require returning `C:\` on Windows
        return "/"

def format_bytes(bytes: int, base: Union[int, None] = None) -> str:
    """ Pretty-print size of bytes. """
    base = base or OTHER_BASE
    for unit in ["", "K", "M", "G", "T", "P"]:
        if bytes < base:
            return f"{bytes:.2f}{unit}B"
        bytes /= base

def get_megabytes(bytes: int, base: Union[int, None] = None) -> int:
    """ Transforms bytes value to megabytes value. """
    base = base or OTHER_BASE
    return bytes / base / base



import psutil

from typing import List

class CPUMonitor(object):
    def __init__(self):
        self.cores = psutil.cpu_count(logical=False)

    def start(self):
        # Provides a meaningless value, 0.0, which should be ignored (per docs)
        psutil.cpu_percent(interval=None)

    def get_stats(self) -> tuple[int, int]:
        """ Get CPU stats.

        @returns number of logical CPU cores and percent utilization
        """
        usage = psutil.cpu_percent(interval=None)
        return self.cores, usage

    def get_formatted_stats(self) -> tuple[int, int]:
        cores, usage = self.get_stats()
        return cores, f"{usage}%"

    def get_values(self, delay: int) -> List[dict]:
        """ Get list of values that represent an `AgentValue`. """
        cores, usage = self.get_stats()
        return [
            {"name": "CPU %", "value": usage}
        ]



import logging
import os
import re

from typing import List, Union
from typing_extensions import Optional, Self

from yaml import load, dump, safe_load
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except:
    from yaml import Loader, Dumper

AVAIL_THRESH_LEVELS = ["warning", "error", "critical"]
AVAIL_STATUS_STATES = ["healthy", "warning", "error", "critical"]
AVAIL_AGENT_TYPES = ["machine", "service", "vendor"]

NODE_RESERVED_NAMES = ["measurements", "c", "xff"]
NODE_VALID_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789_-"
NODE_VALID_FIRST_CHARS = "abcdefghijklmnopqrstuvwxyz"
NODE_MAX_NAME_LENGTH = 30

CONFIG_PATH = None

def get_name() -> str:
    return "ays-agent"

def get_version() -> str:
    return "1.0.0"

def set_config_path(path: str) -> None:
    """ Set the @ys configuration path.

    Use only for testing.
    """
    global CONFIG_PATH
    CONFIG_PATH = path

def get_config_path():
    global CONFIG_PATH
    if CONFIG_PATH:
        return CONFIG_PATH
    home_path = os.path.expanduser("~")
    path = os.path.join(home_path, ".ays-agent")
    return path

def write_yaml(fh, obj):
    dump(obj, fh)

def read_yaml(fh):
    return load(fh, Loader=Loader)

class AgentException(Exception):
    pass

class CLIOptions(object):
    def __init__(
        self,
        org_secret: str,
        server: str,
        parent: str,
        monitor_name: Optional[str] = None,
        interval: Optional[int] = None,
        child: Optional[str] = None,
        create_child: Optional[bool] = None,
        node_type: Optional[str] = None,
        managed: Optional[bool] = None,
        heartbeat_timeout: Optional[int] = None,
        heartbeat_level: Optional[str] = None,
        value: Optional[float] = None,
        value_name: Optional[str] = None,
        value_threshold: Optional[str] = None,
        values: Optional[List[float]] = None,
        value_names: Optional[List[str]] = None,
        value_thresholds: Optional[List[str]] = None,
        status_message: Optional[str] = None,
        status_state: Optional[str] = None,
        monitor_resources: Optional[List[str]] = None,
        monitor_file: Optional[str] = None,
        monitor_program: Optional[str] = None
    ):
        self.org_secret = org_secret
        self.server = server
        self.parent = parent
        self.monitor_name = monitor_name
        self.interval = interval
        self.child = child
        self.create_child = create_child
        self.node_type = node_type
        self.managed = managed
        self.heartbeat_timeout = heartbeat_timeout
        self.heartbeat_level = heartbeat_level
        self.value = value
        self.value_name = value_name
        self.value_threshold = value_threshold
        self.values = values
        self.value_names = value_names
        self.value_thresholds = value_thresholds
        self.status_message = status_message
        self.status_state = status_state
        self.monitor_resources = monitor_resources
        self.monitor_file = monitor_file
        self.monitor_program = monitor_program

        # Options provided to the app from the CLI
        self.cli_options = None

    def setv(self, name, arg) -> None:
        """ Sets the value for Self property `name` from `arg` values. """
        val = arg.get(name, None)
        if val:
            self.__dict__[name] = val

    def get_options(self) -> dict:
        """ Returns all options that can be saved. """
        opts = self.__dict__.copy()
        opts.pop("cli_options", None)
        return opts

    def merge(self, **kwargs) -> None:
        """ Merge existing options with options provided from CLI. """
        self.cli_options = CLIOptions(**kwargs)
        keys = list(self.get_options().keys())
        for key in keys:
            self.setv(key, kwargs)

# Public API

def load_options(path: str) -> CLIOptions:
    """ Load ays agent server options from config file.

    Returns an options, if config file not found.
    """
    if os.path.isfile(path):
        with open(path, "r") as fh:
            opts = read_yaml(fh)
        return CLIOptions(**opts)
    return CLIOptions(
        org_secret="",
        server="",
        parent="",
        monitor_name=""
    )

def save_options(options: CLIOptions) -> None:
    """ Save options to disk. """
    path = get_config_path()
    opts = options.get_options()
    keys = list(opts.keys())
    for key in keys:
        if opts[key] is None:
            del opts[key]
    with open(path, "w") as fh:
        write_yaml(fh, opts)

def get_agent_payload(options) -> None:
    """ Returns a request that represents an `AgentPayload`. """
    if not options.org_secret:
        raise AgentException("'org_secret' must be provided")
    if not options.parent:
        raise AgentException("'parent' must be provided")
    if not options.monitor_name:
        raise AgentException("'monitor_name' must be provided")
    params = {
        "org_secret": options.org_secret,
        "parent": {"property": "path", "value": options.parent},
    }
    if options.child:
        params["relationship"] = {"type": "child", "monitor_name": options.monitor_name, "path": get_formatted_node_name(options.child)}
    elif options.create_child:
        params["relationship"] = {"type": "child", "monitor_name": options.monitor_name, "path": get_formatted_node_name(options.monitor_name.lower())}
    else:
        params["relationship"] = {"type": "parent", "monitor_name": options.monitor_name}
    if options.managed is not None:
        params["managed"] = options.managed
    if options.heartbeat_timeout:
        params["heartbeat"] = {
            "timeout": int(options.heartbeat_timeout),
            "level": get_heartbeat_level(options.heartbeat_level)
        }
    if options.node_type and (options.child or options.create_child):
        params["type"] = get_agent_type(options.node_type.lower())
    if options.value:
        params["value"] = get_value(options.value_name, options.value, options.value_threshold)
    elif options.values:
        params["values"] = get_values(options.value_names, options.values, options.value_thresholds)
    elif options.status_message or options.status_state:
        params["status"] = get_status(options.status_message, options.status_state)
    return options.server, params

# Private API

def get_heartbeat_level(level: str) -> str:
    """ Returns heartbeat level. Returns default if not provided. """
    if level and level not in AVAIL_THRESH_LEVELS:
        raise AgentException(f"Invalid heartbeat level ({level}). Available options are ({', '.join(AVAIL_THRESH_LEVELS)})")
    elif not level:
        level = "critical"
    return level


def get_agent_type(agent_type: str) -> str:
    """ Returns agent type, if valid. """
    if agent_type not in AVAIL_AGENT_TYPES:
        raise AgentException(f"Invalid agent type ({agent_type}). Available options are ({', '.join(AVAIL_AGENT_TYPES)})")
    return agent_type

def get_formatted_node_name(name: str) -> str:
    """ Standardize a node name to follow @ys node name conventions.

    This removes all characters outside of the acceptable range with hyphens.
    Duplicate hyphens are removed.
    """
    if not name:
        raise AgentException("Name must be greater than zero")
    if name[0] not in NODE_VALID_FIRST_CHARS:
        raise AgentException(f"A node name must start with one of the following characters ({NODE_VALID_FIRST_CHARS})")
    if len(name) > NODE_MAX_NAME_LENGTH:
        raise AgentException(f"A node name may not exceed {NODE_MAX_NAME_LENGTH} characters")
    node_name = ""
    for c in name:
        if c not in NODE_VALID_CHARS:
            node_name += "-"
        else:
            node_name += c
    # Remove duplicate hyphens, if any
    node_name = re.sub(r"-+", "-", node_name)
    return node_name

def get_status_state(state) -> str:
    """ Returns provided status state or `critical` if not provided. """
    if state and state not in AVAIL_STATUS_STATES:
        raise AgentException(f"Invalid status state ({state}). Available options are ({', '.join(AVAIL_STATUS_STATES)})")
    elif not state:
        state = "critical"
    return state

def get_status(message, state) -> dict:
    """ Returns a dict that represents an `AgentStatus`. """
    return {
        "message": message or "",
        "state": get_status_state(state)
    }

def get_value(name: Union[str, None], value: str, threshold: Union[str, None], index: Optional[int] = None) -> dict:
    """ Returns a dict value that represents an `AgentValue`. """
    value = {
        "name": name or (index is None and "value" or f"value{index}"),
        "value": float(value),
        "threshold": get_threshold(threshold)
    }
    if not value.get("threshold"):
        value.pop("threshold", None)
    return value

def strip_v(v: str, values: Optional[List[any]] = None):
    if v:
        return list(map(lambda x: x.strip(), v.split(",")))
    elif values is None:
        # NOTE: This fn is a bit overloaded, but it's fine for now. There should probably
        # be 2 strip_v fns. One which requires values, and another that doesn't.
        raise AgentException("Attempting to strip an string value")
    else:
        return [None] * len(values)

def get_values(names: Union[str, None], values: str, thresholds: Union[str, None]) -> List[dict]:
    """ Returns an array of dict values that represent an `AgentValue`. """
    values = strip_v(values, values)
    names = strip_v(names, values)
    if len(names) != len(values):
        raise AgentException(f"The number of names ({len(names)}) must match the number of values ({len(values)}) provided")
    thresholds = strip_v(thresholds, values)
    if len(thresholds) != len(values):
        raise AgentException(f"The number of thresholds ({len(thresholds)}) must match the number of values ({len(values)}) provided")
    r_values = []
    for idx, v in enumerate(values):
        v = get_value(names[idx], v, thresholds[idx], idx)
        r_values.append(v)
    return r_values

def get_threshold_level(level_parts: List[str]) -> str:
    """ Returns a threshold value given threshold parts.

    The first part is ignored. The second part is the threshold value.
    """
    if len(level_parts) > 1:
        level = level_parts[1]
        if level not in AVAIL_THRESH_LEVELS:
            raise AgentException(f"Invalid threshold level ({level}). Available options are ({', '.join(AVAIL_THRESH_LEVELS)})")
    else:
        level = "critical"
    return level

def get_threshold(thresh: str) -> dict:
    """ Returns a dict that represents an `AgentThreshold`. """
    if not thresh:
        return None
    # Apparently the `click` / `typer` library do not remove quotes around values
    thresh = thresh.lower().strip("'").strip('"')

    # Outside threshold range
    if "-" in thresh:
        if ":" in thresh:
            parts = thresh.split(":")
        else:
            parts = [thresh]
        _min, _max = parts[0].split("-")
        level = get_threshold_level(parts)
        return {"outside": {"min": float(_min), "max": float(_max)}, "level": level}

    # Threshold type
    if thresh.startswith(">"):
        thresh_type = "above"
    elif thresh.startswith("<"):
        thresh_type = "below"
    elif thresh.startswith("e"):
        thresh_type = "equal"
    elif thresh.startswith("ne"):
        thresh_type = "nequal"
    else:
        raise AgentException(f"Invalid threshold type for value ({thresh})")

    # Threshold level
    level_parts = thresh.split(":")
    level = get_threshold_level(level_parts)

    # Threshold value
    numeric_part = re.search(r'\d+(\.\d+)?', thresh).group()

    # NOTE: `float` will raise `ValueError` if it's not valid
    return {thresh_type: float(numeric_part), "level": level}



import psutil

from typing import List

from ays_agent.stat import format_bytes

class MemoryMonitor(object):
    def __init__(self):
        self.cores = psutil.cpu_count(logical=False)

    def start(self):
        pass

    def get_stats(self) -> tuple[int, int, int]:
        """ Get memory stats.

        @returns total, used, and percent of memory used
        """
        mem = psutil.virtual_memory()
        return mem.total, mem.total - mem.available, mem.percent

    def get_formatted_stats(self) -> tuple[str, str, str]:
        total, used, percent = self.get_stats()
        return format_bytes(total), format_bytes(used), f"{percent}%"

    def get_values(self, delay: int) -> List[dict]:
        """ Get list of values that represent an `AgentValue`. """
        total, used, percent = self.get_stats()
        return [
            {"name": "RAM %", "value": percent}
        ]



#!/usr/bin/env python3
#
# Setup the ays-agent to be used as a stand-alone executable
#

from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='ays_agent',
    version='1.0',
    authors = [
        {"name": "Eric Chamberlain", "email": "bitheadrl@protonmail.com"}
    ],
    description="@ys Agent Sensor",
    long_description=long_description,
    long_description_content_type="text/markdown",
    readme="README.md",
    requires_python=">=3.8",
    install_requires=[
        "click>=7.1.2",
        "fastapi>=0.109.1",
        "fastapi-utils>=0.2.1",
        "psutil>=5.9.5",
        "PyYaml>=6.0",
        "requests>=2.31.0",
        "rich>=13.7.0",
        "typer>=0.9.0",
        "urllib3>=1.26.18,<2",
        "uvicorn>=0.18.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'ays-agent = ays_agent.cli:app',
        ],
    },
)



import logging
import psutil
import sys

from typing import List
from typing_extensions import Optional

from ays_agent.stat import format_bytes, get_base, get_megabytes, get_default_mountpoint

class DiskMonitor(object):
    def __init__(self, path: Optional[str] = None):
        # Disk path (mountpoint) to monitor
        self.path = path or get_default_mountpoint()
        self.bytes_read = 0
        self.bytes_written = 0

    def start(self):
        """ Start monitoring network traffic. """
        io = psutil.disk_io_counters()
        self.bytes_read, self.bytes_written = io.read_bytes, io.write_bytes

    def get_num_disk(self):
        """ Get the number of physical devices. Ignores pseudo, memory, duplicate, etc. """
        disks = psutil.disk_partitions(all=False)
        return len(disks)

    def get_stats(self, delay: int) -> tuple[int, int, int, int, int, int]:
        """ Get disk stats from last delay.

        @returns total, used, percent, total bytes read, total bytes written, bytes read, bytes written
        """
        usage = psutil.disk_usage(self.path)
        io = psutil.disk_io_counters()
        num_read, num_written = (io.read_bytes - self.bytes_read) / delay, (io.write_bytes - self.bytes_written) / delay
        self.bytes_read, self.bytes_written = io.read_bytes, io.write_bytes
        return usage.total, usage.used, usage.percent, self.bytes_read, self.bytes_written, num_read, num_written

    def get_formatted_stats(self, delay: int) -> tuple[str, str, str, str]:
        """ Get formatted network stats since last delay. """
        total, used, percent, bytes_read, bytes_written, bytes_r, bytes_w = self.get_stats(delay)
        base = get_base()
        def fb(bytes: int) -> str:
            return format_bytes(bytes, base)
        return fb(total), fb(used), f"{percent}%", fb(bytes_read), fb(bytes_written), fb(bytes_r), fb(bytes_w)

    def get_values(self, delay: int) -> List[dict]:
        """ Get list of values that represent an `AgentValue`. """
        total, used, percent, bytes_read, bytes_written, bytes_r, bytes_w = self.get_stats(delay)
        return [
            {"name": "Disk Used %", "value": percent},
            {"name": "Disk R/s", "value": get_megabytes(bytes_r)},
            {"name": "Disk W/s", "value": get_megabytes(bytes_w)}
        ]
